﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Web;

namespace LargeCookieModule
{
    public class CookieHelper
    {
        /// <summary>
        /// firefox.total:4145
        /// chrome.total:4096
        /// </summary>
        private const int CONST_MAX_COOKIE_BYTES = 4096;
        private const string CONST_BEYOND_SEPARATOR = "_C_SPO_";
        /// <summary>
        /// http://www.rfc-editor.org/rfc/rfc6265.txt
        /// At least 4096 bytes per cookie (as measured by the sum of the length of the cookie's name, value, and attributes).
        /// At least 50 cookies per domain.
        /// At least 3000 cookies total.
        /// 目前考虑最多2位数来解决cookie溢出的问题。
        /// 另CONST_BEYOND_SEPARATOR.Length位用来解决SEPARATOR的问题。
        /// </summary>
        private const int CONST_BEYOND_CHARS = (2 + 7);
        private static System.Text.Encoding defaultEncoding
        {
            get
            {
                return System.Text.Encoding.UTF8;
            }
        }

        public static string GetSetCookieHeaderString(HttpContext context, HttpCookie cookie)
        {
            string name = cookie.Name;
            string value = cookie.Value;
            string domain = cookie.Domain;
            DateTime expires = cookie.Expires;
            string path = cookie.Path;
            bool secure = cookie.Secure;
            bool httpOnly = cookie.HttpOnly;

            StringBuilder stringBuilder = new StringBuilder();
            if (!string.IsNullOrEmpty(name))
            {
                stringBuilder.Append(name);
                stringBuilder.Append('=');
            }
            if (!string.IsNullOrEmpty(value))
            {
                stringBuilder.Append(value);
            }
            if (!string.IsNullOrEmpty(domain))
            {
                stringBuilder.Append("; domain=");
                stringBuilder.Append(domain);
            }
            if (expires != DateTime.MinValue)
            {
                stringBuilder.Append("; expires=");
                stringBuilder.Append(FormatHttpCookieDateTime(expires));
            }
            if (!string.IsNullOrEmpty(path))
            {
                stringBuilder.Append("; path=");
                stringBuilder.Append(path);
            }
            if (secure)
            {
                stringBuilder.Append("; secure");
            }
            if (httpOnly && SupportsHttpOnly(context))
            {
                stringBuilder.Append("; HttpOnly");
            }
            return stringBuilder.ToString();
        }

        private static bool SupportsHttpOnly(HttpContext context)
        {
            if (context != null && context.Request != null)
            {
                HttpBrowserCapabilities browser = context.Request.Browser;
                return browser != null && (browser.Type != "IE5" || browser.Platform != "MacPPC");
            }
            return false;
        }

        /// <summary>
        /// equals to HttpUtility.FormatHttpCookieDateTime
        /// for java system to rewrite code.
        /// </summary>
        /// <param name="dt"></param>
        /// <returns></returns>
        private static string FormatHttpCookieDateTime(DateTime dt)
        {
            if (dt < DateTime.MaxValue.AddDays(-1.0) && dt > DateTime.MinValue.AddDays(1.0))
            {
                dt = dt.ToUniversalTime();
            }
            return dt.ToString("ddd, dd-MMM-yyyy HH':'mm':'ss 'GMT'", DateTimeFormatInfo.InvariantInfo);
        }


        /// <summary>
        /// 是否是合法的Cookies（目前只考虑长度）
        /// 目前暂时只知道用System.Text.Encoding.UTF8来解决。无法获取客户端存储的情况。
        /// </summary>
        /// <param name="context"></param>
        /// <param name="cookie"></param>
        /// <returns></returns>
        public static bool IsLengthLegalCookie(HttpContext context, HttpCookie cookie)
        {
            if (cookie == null)
                return false;
            string cookieString = CookieHelper.GetSetCookieHeaderString(context, cookie);

            if (cookieString != null && defaultEncoding.GetByteCount(cookieString) <= CONST_MAX_COOKIE_BYTES)
            {
                return true;
            }
                        
            return false;
        }

        /// <summary>
        /// 获取除了Value（name+value）值以外的字符长度
        /// </summary>
        /// <param name="context"></param>
        /// <param name="cookie"></param>
        /// <returns></returns>
        private static int GetLengthExceptValue(HttpContext context, HttpCookie cookie)
        {
            if (cookie == null)
                return 0;
            string cookieString = CookieHelper.GetSetCookieHeaderString(context, cookie);
            if (!string.IsNullOrEmpty(cookieString))
            {
                int index = cookieString.IndexOf(';');
                // if index equals to -1 result is - (-1) - 1.
                return cookieString.Length - index - 1;
            }
            return 0;
        }

        public static List<HttpCookie> GetCookies(HttpCookieCollection cookies)
        {
            Dictionary<string, HttpCookie> resultCookies = new Dictionary<string, HttpCookie>();
            SeparatorSortedDictionary<HttpCookie> separatorCookies = new SeparatorSortedDictionary<HttpCookie>();

            if (cookies != null && cookies.AllKeys != null && cookies.AllKeys.Length > 0)
            {
                foreach (string key in cookies.AllKeys)
                {
                    HttpCookie cookie = cookies[key];
                    if (cookie != null && !string.IsNullOrEmpty(cookie.Name))
                    {
                        int separatorIndex = cookie.Name.IndexOf(CONST_BEYOND_SEPARATOR);
                        if (separatorIndex != -1)
                        {
                            string prefixName = cookie.Name.Substring(0, separatorIndex);
                            string sOrder = cookie.Name.Substring(prefixName.Length + CONST_BEYOND_SEPARATOR.Length);
                            int iOrder = 0;
                            if (int.TryParse(sOrder, out iOrder))
                            {
                                separatorCookies.Add(prefixName, iOrder, cookie);
                            }
                        }
                        else
                        {
                            resultCookies[cookie.Name] = cookie;
                        }
                    }
                }
                List<string> separatorKeys = separatorCookies.Keys;
                foreach (string key in separatorKeys)
                {
                    if(resultCookies.ContainsKey(key))
                        continue;
                    HttpCookie joinCookie = separatorCookies.Join(key,
                        (c) => c.Value,
                        (s, c) => { c.Value = s; return c; },
                        CloneHttpCookie);
                    resultCookies[key] = joinCookie;
                }
            }
            return resultCookies.Values.ToList();
        }

        public static List<HttpCookie> GetSetCookies(HttpContext context, HttpCookie cookie)
        {
            if (IsLengthLegalCookie(context, cookie))
            {
                List<HttpCookie> cookies = new List<HttpCookie>();
                cookies.Add(cookie);
                return cookies;
            }
            else
            {
                string name = cookie.Name;
                string value = cookie.Value;
                int maxValueCount = CONST_MAX_COOKIE_BYTES - CONST_BEYOND_CHARS - GetLengthExceptValue(context, cookie) - name.Length - 1/*name=value后面的分号*/;
                if (maxValueCount > 0)
                {
                    byte[] bValue = defaultEncoding.GetBytes(value);
                    int iEachChunk = 0, iChunkNum, iChunkCount = 0;
                    SortedList<int, byte[]> chunks = new SortedList<int, byte[]>();
                    iChunkCount = bValue.Length / maxValueCount + 1;
                    int iChunkLastIndex = iChunkCount-1;
                    for (iChunkNum = 0; iChunkNum < iChunkCount; ++iChunkNum)
                    {
                        byte[] chunk = null;
                        if (iChunkNum != iChunkLastIndex)
                        {
                            chunk = new byte[maxValueCount];
                        }
                        else 
                        {
                            chunk = new byte[bValue.Length % maxValueCount];
                        }
                        for (iEachChunk = 0; iEachChunk < chunk.Length; ++iEachChunk)
                        {
                            chunk[iEachChunk] = bValue[iChunkNum * maxValueCount + iEachChunk];
                        }
                        chunks.Add(iChunkNum, chunk);
                    }
                    List<HttpCookie> cookies = new List<HttpCookie>();
                    foreach (KeyValuePair<int, byte[]> item in chunks)
                    {
                        string itemName = name + CONST_BEYOND_SEPARATOR + item.Key.ToString();
                        string itemValue = defaultEncoding.GetString(item.Value);
                        HttpCookie cloneCookie = CloneHttpCookie(cookie);
                        cloneCookie.Name = itemName;
                        cloneCookie.Value = itemValue;
                        cookies.Add(cloneCookie);
                    }
                    return cookies;
                }
            }
            return null;
        }

        public static bool IsSeparatorCookie(HttpCookie cookie)
        {
            if (cookie == null)
                return false;
            if (!string.IsNullOrEmpty(cookie.Name) && cookie.Name.Contains(CONST_BEYOND_SEPARATOR))
            {
                return true;
            }
            return false;
        }

        private static HttpCookie CloneHttpCookie(HttpCookie cookie)
        {
            HttpCookie cloneCookie = new HttpCookie(cookie.Name);
            cloneCookie.Value = cookie.Value;
            cloneCookie.Path = cookie.Path;
            cloneCookie.Secure = cookie.Secure;
            cloneCookie.HttpOnly = cookie.HttpOnly;
            cloneCookie.Domain = cookie.Domain;
            cloneCookie.Expires = cookie.Expires;
            return cloneCookie;
        }

        public static void FilterRequestCookies(HttpContext context)
        {
            if (context != null)
            {
                IList<HttpCookie> cookies = CookieHelper.GetCookies(context.Request.Cookies);
                if (cookies != null)
                {
                    foreach (HttpCookie subCookie in cookies)
                    {
                        if (context.Request.Cookies[subCookie.Name] == null)
                        {
                            context.Request.Cookies.Add(subCookie);
                        }
                    }
                }
            }
        }

        public static void FilterResponseCookies(HttpContext context)
        {
            if (context != null && context.Response.Cookies != null && context.Response.Cookies.Keys.Count > 0)
            {
                int keyCount = context.Response.Cookies.Keys.Count;
                if (keyCount > 0)
                {
                    string[] keys = new string[keyCount];
                    for (int i = 0; i < keyCount; ++i)
                    {
                        keys[i] = context.Response.Cookies.Keys[i];
                    }

                    foreach (string key in keys)
                    {
                        HttpCookie everyCookie = context.Response.Cookies[key];

                        IList<HttpCookie> cookies = CookieHelper.GetSetCookies(context, everyCookie);
                        if (cookies != null && cookies.Count > 1 /*means it is separator cookie or the raw cookie is not legal in length.*/)
                        {
                            foreach (HttpCookie subCookie in cookies)
                            {
                                if (IsSeparatorCookie(subCookie))
                                {
                                    context.Response.AppendCookie(subCookie);
                                }
                            }
                        }
                    }
                }
            }
        }

        internal class SeparatorSortedDictionary<T> : Dictionary<string, SortedList<int, T>>
        {
            Dictionary<string, SortedList<int, T>> _innerStore = new Dictionary<string, SortedList<int, T>>();
            public void Add(string name, int index, T value)
            {
                if (_innerStore != null)
                {
                    if (!_innerStore.ContainsKey(name))
                        _innerStore[name] = new SortedList<int, T>();
                    SortedList<int, T> itemCollection = _innerStore[name];
                    if (!itemCollection.ContainsKey(index))
                    {
                        itemCollection.Add(index, value);
                    }
                }
            }

            public T Join(string name, Func<T, string> toStringFunc, Func<string, T, T> fromStringFunc, Func<T, T> cloneFunc)
            {
                if (_innerStore != null)
                {
                    if (_innerStore.ContainsKey(name))
                    {
                        SortedList<int, T> itemCollection = _innerStore[name];
                        string value = string.Empty;
                        T firstElement = default(T);
                        bool getFirstElement = false;
                        foreach (T itemValue in itemCollection.Values)
                        {
                            if (!getFirstElement)
                            {
                                firstElement = itemValue;
                                getFirstElement = true;
                            }
                            value += toStringFunc(itemValue);
                        }
                        T cloneItem = cloneFunc(firstElement);
                        cloneItem = fromStringFunc(value, cloneItem);
                        return cloneItem;
                    }
                }
                return default(T);
            }

            public List<string> Keys
            {
                get
                {
                    List<string> result = new List<string>();
                    if (_innerStore != null)
                    {
                        foreach (string key in _innerStore.Keys)
                        {
                            result.Add(key);
                        }
                    }
                    return result;
                }
            }
        }
    }
}
